/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* Graph class                                                     */
/* contains a matrix representation of a graph                     */
/* with a list of subtrees and a list of automorphisms             */

package Homology;

import java.util.*;

public class Graph extends Object implements java.io.Serializable
{
  public int size;			// number of vertices
  public byte m[][];		// to store the matrix 
  public int valency;		// of the base point
  public int neighbors;		// of the base point; <= valency
  public int noEdges;		// # of edges (double edges count as 1)
  public Vector treeBag;	// Prfer codes of the max subtrees
  public Vector symBag;		// the automorphisms of the graph
  
  public static final int[] bagSize = {0,0,10,10,10,10,10,50,100,500};
  						// ballpark tree counts for the various degrees
  
  public Graph(int n) {			// empty graph
    size = n;
    m = new byte[size][size];
    treeBag = new Vector(bagSize[n-1],bagSize[n-1]);
    symBag = new Vector(5,5);
    }
   
   
  public Graph(Graph g) {		// return a clone of the Graph
    this(g.size);
    int i,j;
    
    for (i=0; i<size; i++)		// only storing #s above the diagonal
      for (j=i+1; j<size; j++)	//  (there are no loop-edges)
        m[i][j] = g.m[i][j];
    valency = g.valency;
    neighbors = g.neighbors;
    noEdges = g.noEdges;
    }
    

  public Graph permuted(Permutation sigma) {
    int i, j, x, y, t;					// sigma is a permutation
    Graph g = new Graph(size);			// of 'size' numbers,
    									// with sigma.p[0] = 0
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++) {
        x = sigma.p[i];			// permute the labels on the vertices
        y = sigma.p[j];			//   by sigma.inverse()
        if (x > y)
          { t=x; x=y; y=t; }
        g.m[i][j] = m[x][y];
        }				
    g.valency = valency;
    g.neighbors = neighbors;
    g.noEdges = noEdges;
    return g;
    }
    
    
  public boolean equals(Graph otherg) {		// check whether the adj.
    int i, j;								//   matrices are equal
    
    for (i=0; i<size; i++)			// not checking size: this is only
      for (j=i+1; j<size; j++)		// called when they are the same
        if (m[i][j] != otherg.m[i][j])
          return false;
    return true;
    }
    
    
  public boolean isIsomorphic(Graph otherg) {	// not really checking
  							// whether the graphs are isomorphic; we are
    int i, j;				// assuming they are both in normal form ...
    
    if (size != otherg.size || valency != otherg.valency ||
        neighbors != otherg.neighbors || noEdges != otherg.noEdges)
      return false;

    return equals(otherg);	// ... and then whether they are equal
    }
    

  public static int order(Graph g1, Graph g2) {
    int i, j;			
    								// return 0 if the graphs are equal,
    for (i=0; i<g1.size; i++)		// -1 if g1 < g2 and 1 if g1 > g2
      for (j=i+1; j<g1.size; j++) {
        if (g1.m[i][j] < g2.m[i][j])
          return -1;
        if (g1.m[i][j] > g2.m[i][j])
          return 1;
        }
    return 0;
    }


  public void toNormalForm() {
    int i, j;					// put graph in normal form
    Graph g, gmax;				//  and find automorphisms
    Permutation sigma, sstored;
    Vector tempSymBag = new Vector(5,5);

    sigma = new Permutation(size);
    sstored = new Permutation(size);
    gmax = new Graph(this);

    do {
      switch (order((g = this.permuted(sigma)), gmax)) {
     								// found a new automorphism
        case 0:  tempSymBag.addElement(new Permutation(sigma));
                 break;
                 
        case 1:  gmax = g;			// found a new lexicographic max
                 sstored = new Permutation(sigma);
                 tempSymBag.removeAllElements();
                 tempSymBag.addElement(sstored);
                 break;
        }
      sigma.next();
      } while (sigma.p[0] == 0);
    
    for (i=0; i<size; i++)		
      for (j=i+1; j<size; j++)
        m[i][j] = gmax.m[i][j];
    
    sstored = sstored.inverse();
    for (i=0; i<tempSymBag.size(); i++) {
      sigma = (Permutation)tempSymBag.elementAt(i);
      symBag.addElement(sstored.compose(sigma));
      }
    }

      
  public Graph plusedge(int edge) {		// add a new edge between the
    int i, j, k;						// basepoint and the midpoint
    Graph g = new Graph(size + 1);		// of an edge at the basepoint
    
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        g.m[i][j] = m[i][j];	// copy old graph into 'g'

    j=0; k=0;
    do {
      j++;
      if (m[0][j] > 0)
        k++;
      } while (k <= edge);		// find edge number 'edge'
    
    g.m[0][j]--;
    g.m[0][size] = 2;
    g.m[j][size] = 1;			// modify the adjacency matrix
    g.valency = valency + 1;	
    if (g.m[0][j] == 0) {		
      g.neighbors = neighbors;
      g.noEdges = noEdges + 1;
      }
    else {						// fix the statistics of the graph
      g.neighbors = neighbors + 1;
      g.noEdges = noEdges + 2;
      }
    return g;
    }
    
    
  public Graph glueedge(int edge1, int edge2) {	
    int i, j, k;						// glue the initial half
    Graph g = new Graph(size + 1);		// of two edges at the 
    									// basepoint together
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        g.m[i][j] = m[i][j];	// copy old graph into 'g'

    i=0; k=0;
    do {
      i++;
      if (m[0][i] > 0)
        k++;
      } while (k <= edge1);		// find 'edge1' edge

    j=i; 
    while (k <= edge2) {
      j++;
      if (m[0][j] > 0)
        k++;
      }					 		// find 'edge2' edge
    
    g.m[0][i]--;
    g.m[0][j]--;
    g.m[0][size] = 1;
    g.m[i][size] = 1;
    g.m[j][size] = 1;			// modify matrix
    g.valency = valency - 1;	// (we are assuming edge1 != edge2)
    k = 0;
    if (g.m[0][i] != 0)  k++;
    if (g.m[0][j] != 0)  k++;
    g.neighbors = neighbors - 1 + k;
    g.noEdges = noEdges + 1 + k;
    return g;
    }
    
    
  public Tree subGraph(int[] subset) {	// construct a subtree of 'this'
    int i, j, k;
    Tree t = new Tree(size);
    
    k = -1;
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        if (m[i][j] != 0) {
          k++;
          if (subset[k] != 0)
            t.m[i][j] = 1;
          }
    t.noEdges = k+1;
    t.parent = this;
    return t;
    }
}
