/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* Tree class                                                      */
/* Class for storing a tree on n numbered vertices.                */

package Homology;

import java.util.*;

public class Tree extends Object implements java.io.Serializable
{
  
  public int size;				// number of vertices
  public byte m[][];			// to store the matrix 
  public int noEdges, coeff;	// number of edges; coeff is plus or minus 1
  public Graph parent;			// points to ambient graph
  
  public Tree(int n) {		// empty tree
    size = n;
    m = new byte[size][size];
    coeff = 1;
    }
 

  public Tree(Graph g) {	// construct a "subtree" of a graph
    this(g.size);			// (not really a tree: just make all
    int i, j;				// multiple edges single, and make it
    						// of type Tree)
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++) {
        if (g.m[i][j] != 0)
          m[i][j] = 1;
        else
          m[i][j] = 0;
        }
    noEdges = g.noEdges;
    parent = g;
    }


  public Tree(Tree t) {		// a clone of 'this' Tree
    this(t.size);
    int i, j;

    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        m[i][j] = t.m[i][j];
    noEdges = t.noEdges;
    parent = t.parent;
    }


  public Tree(byte prf[]) {		// construct a Tree from its Prfer code
    this(prf.length + 2);
    
    int i, j, min;
    int p[] = new int[size-1];
    int q[] = new int[size-1];
    boolean used[] = new boolean[size];
   
    for (i=0; i<size-2; i++)
      p[i] = prf[i];
    p[size-2] = size-1;			// put an 'n' at the end of the Prfer code
    							// n = label of the last vertex = 'size' - 1
    for (i=0; i<size-1; i++) {
      for (j=0; j<size-1; j++)
        used[j] = false;
      for (j=0; j<i; j++)
        used[q[j]] = true;		// q[i] is the smallest number which does not
      for (j=i; j<size-1; j++)	// occur among q[0], ... q[i-1], p[i], ... p[n-1]
        used[p[j]] = true;
      j = 0;
      while (used[j])
        j++;
      q[i] = j;					// the edges of the tree connect p[i] and q[i]
      }							//   for 0 <= i < size-1

    for (i=0; i<size-1; i++) {
      int x = p[i];
      int y = q[i];
      if (x > y) {
        int tmp=x; x=y; y=tmp;
        }
      m[x][y] = 1;				// construct the adjacency matrix
      }
    noEdges = size - 1;
    }


  public byte[] prfer() {		// find the Prfer code of 'this' Tree
    int n, i, j, k;
    int buffer[][] = new int[size][size];
    byte pr[] = new byte[size-2];
    
    for (i=0; i<size; i++)		// the whole (symmetric) adj. matrix
      for (j=i+1; j<size; j++)
        buffer[i][j] = buffer[j][i] = (int)m[i][j];
    
    for (n=0; n<size-2; n++) {
    
      i = -1;
      do {
        i++;
        k = 0;					// find the first vertex of valence one
        for (j=0; j<size; j++)
          if (buffer[i][j] != 0)
            k++;
        } while (k != 1);
        
      for (j=0; j<size && buffer[i][j] == 0; j++);
      
      pr[n] = (byte) j;			// write down the label of its neighbor ...
      buffer[i][j] = buffer[j][i] = 0;	// ... and erase the edge
      }
    return pr;
    }
        

  public Tree permuted(Permutation sigma) {
    int i, j;							// sigma is a permutation
    Tree t = new Tree(size);			// of 'size' numbers,
    									// with sigma.p[0] = 0
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++) {
        int x = sigma.p[i];
        int y = sigma.p[j];
        if (x > y) 						// permute the labels on the
          { int tmp=x; x=y; y=tmp; }	// vertices by sigma.inverse()
        t.m[i][j] = m[x][y];
        }
    t.noEdges = noEdges;
    t.parent = parent;
    t.coeff = coeff * sign(sigma);
    return t;
    }
    
    
  public int sign(Permutation sigma) {
    int i, j, k;				// computes the sign of the
    Tree t = new Tree(size);	// permutation (on the edges)
    Permutation s = new Permutation(noEdges);
				    			// which takes this.permuted(sigma)
    k = 1;						// to its normal form		
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++) {		// first number the edges
        if (m[i][j] != 0)				// of 'this' tree ...
          t.m[i][j] = (byte)k++;
        }
        
    k = 0;
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++) {		// ... then record the order
        int x = sigma.p[i];				// in which the edges come in
        int y = sigma.p[j];				// the permuted tree ...
        if (x > y) 
          { int tmp=x; x=y; y=tmp; }
        if (t.m[x][y] != 0)
          s.p[k++] = t.m[x][y] - 1;
        }
        
    return s.sign();			// ... and check whether it's odd or even
    }


  public boolean equals(Tree t) {	// checks whether the adj. matrices
    int i, j;						//   are equal (on the nose)
    
    if (noEdges != t.noEdges)
      return false;
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        if (m[i][j] != t.m[i][j])
          return false;
    return true;
    }


  public boolean isConnected() {	// checks whether the Tree is connected
    int i, j, k;
    int tmp[][];
    
    tmp = new int[size][size];
        
    for (i=0; i<size; i++)			// copy the adjacency matrix to a buffer
      for (j=i; j<size; j++)
        tmp[i][j] = (int)m[i][j];
 
    for (i=0; i<size-1; i++) {		// Start with vertex 0; if it is not
      for (j=i+1; j<size; j++)		// connected to anything, return 'false'.
        if (tmp[i][j] != 0)			
          break;					// If 0 is connected to j, collapse the
      if (j == size)				// edge between 0 and j (and call the
        return false;				// new vertex j).
      for (k=j+1; k<size; k++)
        tmp[j][k] |= tmp[i][k];		// Continue: check whether vertex 1 is
      }								// connected to anything, etc.
    return true;
    }						// return 'true' if we never returned 'false'
   
  
  public Tree removeEdge(int edge) {
    int i, j, k;						// remove and edge from the Tree
    Tree t = new Tree(size);    
    
    k = -1;
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        if (m[i][j] != 0) {
          k++;
          if (k!=edge)
            t.m[i][j] = 1;
          }
    t.noEdges = noEdges - 1;
    t.parent = parent;
    return t;
    }     
    
  
  private static void justContract(byte[][] mfrom, byte[][] mto, 
                                                   int x, int y) {
    int i, j;
    								// this is the nitty-gritty of
    for (i=0; i<x; i++) {			//  contracting an edge of a tree
      for (j=i+1; j<x; j++)
        mto[i][j] = mfrom[i][j];
      mto[i][x] = (byte)(mfrom[i][x] + mfrom[i][y]);
      for (j=x+1; j<y; j++)
        mto[i][j] = mfrom[i][j];
      for (j=y+1; j<mfrom.length; j++)
        mto[i][j-1] = mfrom[i][j];
      }
      
    for (j=x+1; j<y; j++)
      mto[x][j] = (byte)(mfrom[x][j] + mfrom[j][y]);
    for (j=y+1; j<mfrom.length; j++)
      mto[x][j-1] = (byte)(mfrom[x][j] + mfrom[y][j]);
      
    for (i=x+1; i<y; i++) {
      for (j=i+1; j<y; j++)
        mto[i][j] = mfrom[i][j];
      for  (j=y+1; j<mfrom.length; j++)
        mto[i][j-1] = mfrom[i][j];
      }
      
    for (i=y+1; i<mfrom.length; i++)
      for (j=i+1; j<mfrom.length; j++)
        mto[i-1][j-1] = mfrom[i][j];
    }
    
  
  public Tree contractEdge(int which, boolean atBasepoint) {
    int i, j = 0, k, p, q;				// contract an edge of the Tree
    Graph g = new Graph(size - 1);
    Tree t = new Tree(size - 1);
    
    k = -1;
i_loop: for (i=0; i<size; i++)			// find the edge to be
          for (j=i+1; j<size; j++) {	//   contracted
            if (m[i][j] != 0)
              k++;
            if (k == which)
              break i_loop;
            }
    
    if (i==0 && atBasepoint || i!=0 && !atBasepoint) {
      justContract(parent.m, g.m, i, j);	// construct the new parent
      for (q=1; q<g.size; q++) {
        if (g.m[0][q] != 0) {
          g.neighbors++;
          g.valency += g.m[0][q];
          }
        }
      g.noEdges = g.neighbors;			// there must be a smarter way
      for (p=1; p<g.size; p++)			// to count the edges, but the
        for (q=p+1; q<g.size; q++)		// dumb way is good enough for us
          if (g.m[p][q] != 0)
            g.noEdges++;

      justContract(m, t.m, i, j);		// construct the new tree
      t.noEdges = noEdges - 1;
      t.parent = g;
      t.coeff = coeff * contractSign(i, j);
      return t;
      }
    else
      return null;
    }  
    

  public int contractSign(int i, int j) {
    int sign, c1, c2,            // compute the sign of the
        p, q, r;                 // contracted tree
                 
    sign = c1 = c2 = 0;
    for (p = 0; p < i; p++)           // count inversions of the form
      if (m[p][j] != 0) {             // (x, j) -> (x, i),  x < i
        for (q = i+1; q < j; q++)
          if (m[p][q] != 0)
            sign++;
        }

    for (p = i+1; p < j; p++)        // count inversions of the form
      if (m[p][j] != 0) {            // (x, j) -> (i, x),  i < x < j
        for (q = p+1; q < size; q++)
          if (q != j && m[i][q] != 0)
            sign++;
        for (q = i+1; q < p; q++)
          for (r = q+1; r < size; r++)
            if (r != j && m[q][r] != 0)
              sign++;
        for (q = p+1; q < j; q++)
          if (m[p][q] != 0)
            sign++;
        }

    for (p = j+1; p < size; p++)      // count inversions of the form
      if (m[j][p] != 0) {             // (j, x) -> (i, x),  x > j
        c1++;
        for (q = p+1; q < size; q++)
          if (m[i][q] != 0)
            sign++;
        }
    if (c1 != 0) {
      for (p = i+1; p < j; p++)
        for (q = p+1; q < size; q++)
          if (q != j && m[p][q] != 0)
            c2++;
      sign += c1 * c2;
      }

    return (sign%2 == 0)? 1 : -1;
    } 


  public Tree normalForm(boolean needsSyms) {
    int i, j;						// Put Tree in normal form
    Graph g, gmax;					// and find automorphisms if needed.
    Permutation sigma, sstored;		// This is a duplication
    Vector tempSymBag; 				// of the toNormalForm()
	Tree t;							// method of Homology.Graph
									// (but its syntax is different).
    sigma = new Permutation(size);
    sstored = new Permutation(size);
    gmax = new Graph(parent);
    if (needsSyms)
      tempSymBag = new Vector(5,5);
    else
      tempSymBag = null;

    do {
      switch (Graph.order((g = parent.permuted(sigma)), gmax)) {
     
        case 0:  if (needsSyms)     // found a new automorphism
                   tempSymBag.addElement(new Permutation(sigma));
                 break;	
                 			
        case 1:  gmax = g;			// found a new lexicographic max
                 sstored = new Permutation(sigma);
                 if (needsSyms) {
                   tempSymBag.removeAllElements();
                   tempSymBag.addElement(sstored);
                   }
                 break;
        }
      sigma.next();
      } while (sigma.p[0] == 0);
    
    t = this.permuted(sstored);
    t.parent = gmax;       

    if (needsSyms) {
      sstored = sstored.inverse();
      for (i=0; i<tempSymBag.size(); i++) {
        sigma = (Permutation)tempSymBag.elementAt(i);
        t.parent.symBag.addElement(sstored.compose(sigma));
        }
      }

    return t;
    }

      
  public boolean isIsomorphic(Tree otherTree) {
    int i;						// assuming the Trees are in normal form, we
    Permutation sigma;			// just check whether their matrices are equal
    
    for (i=0; i<parent.symBag.size(); i++) {
      sigma = (Permutation)parent.symBag.elementAt(i);
      if (otherTree.equals(this.permuted(sigma)))
        return true;
      }
    return false;
    }
    
    
  public Permutation findIsomorphism(Tree otherTree) {
    int i;						// same as isIsomorphic, except returns the
    Permutation sigma;			// permutation which takes 'this' to 'otherTree'
    
    for (i=0; i<parent.symBag.size(); i++) {
      sigma = (Permutation)parent.symBag.elementAt(i);
      if (otherTree.equals(this.permuted(sigma)))
        return sigma;
      }
    return null;				// (or 'null' if there isn't one)
    }
    
        
  public boolean hasOddSymmetry() {		// checks whether the cube
    int i, j;					// ['this.parent', 'this'] has an orientation-
    Permutation sigma;			// reversing automorphism
    							
    for (i=0; i<parent.symBag.size(); i++) {
      sigma = (Permutation)parent.symBag.elementAt(i);
      if (equals(permuted(sigma)) && sign(sigma) == -1)
        return true;
      }
    return false;
    }
    
        
  public boolean containsBadTriangle() {	// checks whether the cube
    int i, j, k;	// falls under one of the lemmas of Hatcher--Vogtmann's
					// Rational homology of Aut(F_n); hence contractible
    								
    for (i=1; i<size; i++)			// Hatcher--Vogtmann Lemma 4.4
      for (j=i+1; j<size; j++)
        if (parent.m[i][j] > 0 && parent.m[0][i] > 1 
                               && parent.m[0][j] == 1) {
          if (m[i][j] > 0 && m[0][i] == 0 && m[0][j] == 0)
            return true;
          else if (m[i][j] == 0 && m[0][i] > 0 && m[0][j] == 0)
            return true;
          }
    for (i=1; i<size; i++)			// Hatcher--Vogtmann Lemma 4.3
      for (j=i+1; j<size; j++)
        for (k=j+1; k<size; k++)
          if (parent.m[i][j] > 0 && parent.m[i][k] > 0
                                 && parent.m[j][k] > 0) {
            if (m[i][j] > 0 && m[i][k] == 0 && m[j][k] == 0)
              return true;
            else if (m[i][j] == 0 && m[i][k] > 0 && m[j][k] == 0)
              return true;
            else if (m[i][j] == 0 && m[i][k] == 0 && m[j][k] > 0)
              return true;          
            }
    return false;
    }
    
}
